#!/usr/bin/perl -w
# NAME: cntntypes.pl
# AIM: *** VERY SPECIFIC *** Read a large json all.all.json file, and enumerate the objects
use strict;
use warnings;
use File::Basename;  # split path ($name,$dir,$ext) = fileparse($file [, qr/\.[^.]*/] )
use JSON;
use Data::Dumper;
use Time::gmtime;
use Cwd;
my $os = $^O;
my $perl_dir = '/home/geoff/bin';
my $PATH_SEP = '/';
my $temp_dir = '/tmp';
if ($os =~ /win/i) {
    $perl_dir = 'C:\GTools\perl';
    $temp_dir = $perl_dir;
    $PATH_SEP = "\\";
}
unshift(@INC, $perl_dir);
require 'lib_utils.pl' or die "Unable to load 'lib_utils.pl' Check paths in \@INC...\n";
# log file stuff
our ($LF);
my $pgmname = $0;
if ($pgmname =~ /(\\|\/)/) {
    my @tmpsp = split(/(\\|\/)/,$pgmname);
    $pgmname = $tmpsp[-1];
}
my $outfile = $temp_dir.$PATH_SEP."temp.$pgmname.txt";
open_log($outfile);

# user variables
my $VERS = "0.0.2 2014-02-22";
my $load_log = 0;
my $in_file = 'C:\OSGeo4W\apache\htdocs\map-test2\SIDSTAR\all.all.json';
my $verbosity = 0;
my $out_file = '';

# ### DEBUG ###
my $debug_on = 0;
my $def_file = 'def_file';

### program variables
my @warnings = ();
my $cwd = cwd();

sub VERB1() { return $verbosity >= 1; }
sub VERB2() { return $verbosity >= 2; }
sub VERB5() { return $verbosity >= 5; }
sub VERB9() { return $verbosity >= 9; }

sub show_warnings($) {
    my ($val) = @_;
    if (@warnings) {
        prt( "\nGot ".scalar @warnings." WARNINGS...\n" );
        foreach my $itm (@warnings) {
           prt("$itm\n");
        }
        prt("\n");
    } else {
        prt( "\nNo warnings issued.\n\n" ) if (VERB9());
    }
}

sub pgm_exit($$) {
    my ($val,$msg) = @_;
    if (length($msg)) {
        $msg .= "\n" if (!($msg =~ /\n$/));
        prt($msg);
    }
    show_warnings($val);
    close_log($outfile,$load_log);
    exit($val);
}


sub prtw($) {
   my ($tx) = shift;
   $tx =~ s/\n$//;
   prt("$tx\n");
   push(@warnings,$tx);
}


sub process_in_file($) {
    my ($inf) = @_;
    prt("Loading [$inf]...\n");
    if (! open INF, "<$inf") {
        pgm_exit(1,"ERROR: Unable to open file [$inf]\n"); 
    }
    my @lines = <INF>;
    close INF;
    my $lncnt = scalar @lines;
    prt("Processing $lncnt lines, from [$inf]...\n");
    my ($line,$rrows,$lnn);
    $line = join("",@lines);
    my $json = JSON->new->allow_nonref;
    my $perl_scalar = $json->decode( $line );
    $lnn = ${$perl_scalar}{row_count};
    prt("Row count $lnn ");
    $rrows = ${$perl_scalar}{rows};
    $lnn = scalar @{$rrows};
    prt("($lnn)\n");
    my ($i,$row,$typ,$id,$lat,$lon,$alt,$nam);
    my $aptcnt = 0; #	Airport
    my $vorcnt = 0; #	VOR Station
    my $dmecnt = 0; #	DME Station
    my $fixcnt = 0; #	a fix
    my $rwycnt = 0; #	Runway
    my $thrcnt = 0; #	Threshold
    my $imcnt = 0;  #	Inner Marker
    my $mmcnt = 0;  #	Middle Marker
    my $omcnt = 0;  #	Outer Marker
    my $ndbcnt = 0; #
    my $othcnt = 0;
    for ($i = 0; $i < $lnn; $i++) {
        $row = ${$rrows}[$i];
        $typ = ${$row}{ntype};
        $lat = ${$row}{lat};
        $lon = ${$row}{lon};
        $alt = ${$row}{elev_m};
        $id  = ${$row}{ident};
        $nam = ${$row}{name};
        prt("$typ,$id,$lat,$lon,$alt\n");
        if ($typ eq 'apt') { 
            $aptcnt++;  # Airport
        } elsif ($typ eq 'vor') { 
            $vorcnt++;  #	VOR Station
        } elsif ($typ eq 'dme') { 
            $dmecnt++; #	DME Station
        } elsif ($typ eq 'fix') { 
            $fixcnt++;  #	a fix
        } elsif ($typ eq 'rwy') { 
            $rwycnt++;  #	Runway
        } elsif ($typ eq 'thr') { 
            $thrcnt++;  #	Threshold
        } elsif ($typ eq 'im') { 
            $imcnt++;   #	Inner Marker
        } elsif ($typ eq 'mm') { 
            $mmcnt++;   #	Middle Marker
        } elsif ($typ eq 'om') { 
            $omcnt++;   #	Outer Marker
        } elsif ($typ eq 'ndb') { 
            $ndbcnt++;
        } else {
            $othcnt++;
        }
    }
    prt("$aptcnt Apt ");
    prt("$vorcnt VOR ");
    prt("$dmecnt DME ");
    prt("$fixcnt fix ");
    prt("$rwycnt rwy ");
    prt("$thrcnt Thr ");
    prt("$imcnt  IM  ");
    prt("$mmcnt  MM  ");  #	Middle Marker
    prt("$omcnt  OM  ");  #	Outer Marker
    prt("$ndbcnt ndb "); #
    prt("$othcnt OTH ");
    prt("\n");
}

#########################################
### MAIN ###
#parse_args(@ARGV);
process_in_file($in_file);
pgm_exit(0,"");
########################################

sub need_arg {
    my ($arg,@av) = @_;
    pgm_exit(1,"ERROR: [$arg] must have a following argument!\n") if (!@av);
}

sub parse_args {
    my (@av) = @_;
    my ($arg,$sarg);
    my $verb = VERB2();
    while (@av) {
        $arg = $av[0];
        if ($arg =~ /^-/) {
            $sarg = substr($arg,1);
            $sarg = substr($sarg,1) while ($sarg =~ /^-/);
            if (($sarg =~ /^h/i)||($sarg eq '?')) {
                give_help();
                pgm_exit(0,"Help exit(0)");
            } elsif ($sarg =~ /^v/) {
                if ($sarg =~ /^v.*(\d+)$/) {
                    $verbosity = $1;
                } else {
                    while ($sarg =~ /^v/) {
                        $verbosity++;
                        $sarg = substr($sarg,1);
                    }
                }
                $verb = VERB2();
                prt("Verbosity = $verbosity\n") if ($verb);
            } elsif ($sarg =~ /^l/) {
                if ($sarg =~ /^ll/) {
                    $load_log = 2;
                } else {
                    $load_log = 1;
                }
                prt("Set to load log at end. ($load_log)\n") if ($verb);
            } elsif ($sarg =~ /^o/) {
                need_arg(@av);
                shift @av;
                $sarg = $av[0];
                $out_file = $sarg;
                prt("Set out file to [$out_file].\n") if ($verb);
            } else {
                pgm_exit(1,"ERROR: Invalid argument [$arg]! Try -?\n");
            }
        } else {
            $in_file = $arg;
            prt("Set input to [$in_file]\n") if ($verb);
        }
        shift @av;
    }

    if ($debug_on) {
        prtw("WARNING: DEBUG is ON!\n");
        if (length($in_file) ==  0) {
            $in_file = $def_file;
            prt("Set DEFAULT input to [$in_file]\n");
        }
    }
    if (length($in_file) ==  0) {
        pgm_exit(1,"ERROR: No input files found in command!\n");
    }
    if (! -f $in_file) {
        pgm_exit(1,"ERROR: Unable to find in file [$in_file]! Check name, location...\n");
    }
}

sub give_help {
    prt("$pgmname: version $VERS\n");
    prt("Usage: $pgmname [options] in-file\n");
    prt("Options:\n");
    prt(" --help  (-h or -?) = This help, and exit 0.\n");
    prt(" --verb[n]     (-v) = Bump [or set] verbosity. def=$verbosity\n");
    prt(" --load        (-l) = Load LOG at end. ($outfile)\n");
    prt(" --out <file>  (-o) = Write output to this file.\n");
}

# eof - template.pl
